#!/usr/bin/env python3

#
# This file is part of KDToolBox.
#
# SPDX-FileCopyrightText: 2022 Klarälvdalens Datakonsult AB, a KDAB Group company <info@kdab.com>
# Author: Jesper K. Pedersen <jesper.pedersen@kdab.com>
#
# SPDX-License-Identifier: MIT
#

'''
Script to add inclusion of mocs to files recursively.
'''

# pylint: disable=redefined-outer-name

import os
import re
import argparse
import sys

dirty = False


def stripInitialSlash(path):
    if path and path.startswith("/"):
        path = path[1:]
    return path

# Returns true if the path is to be excluded from the search


def shouldExclude(root, path):
    # pylint: disable=used-before-assignment
    if not args.excludes:
        return False  # No excludes provided

    assert root.startswith(args.root)
    root = stripInitialSlash(root[len(args.root):])

    if args.headerPrefix:
        assert root.startswith(args.headerPrefix)
        root = stripInitialSlash(root[len(args.headerPrefix):])

    return (path in args.excludes) or (root + "/" + path in args.excludes)


regexp = re.compile("\\s*(Q_OBJECT|Q_GADGET|Q_NAMESPACE)\\s*")
# Returns true if the header file provides contains a Q_OBJECT, Q_GADGET or Q_NAMESPACE macro


def hasMacro(fileName):
    with open(fileName, "r", encoding="ISO-8859-1") as fileHandle:
        for line in fileHandle:
            if regexp.match(line):
                return True
        return False

# returns the matching .cpp file for the given .h file


def matchingCPPFile(root, fileName):
    assert root.startswith(args.root)
    root = stripInitialSlash(root[len(args.root):])

    if args.headerPrefix:
        assert root.startswith(args.headerPrefix)
        root = stripInitialSlash(root[len(args.headerPrefix):])

    if args.sourcePrefix:
        root = args.sourcePrefix + "/" + root

    return args.root + "/"  \
        + root + ("/" if root != "" else "") \
        + fileNameWithoutExtension(fileName) + ".cpp"


def fileNameWithoutExtension(fileName):
    return os.path.splitext(os.path.basename(fileName))[0]

# returns true if the specifies .cpp file already has the proper include


def cppHasMOCInclude(fileName):
    includeStatement = '#include "moc_%s.cpp"' % fileNameWithoutExtension(fileName)
    with open(fileName, encoding="utf8") as fileHandle:
        return includeStatement in fileHandle.read()


def getMocInsertionLocation(filename, content):
    headerIncludeRegex = re.compile(r'#include "%s\.h".*\n' % fileNameWithoutExtension(filename), re.M)
    match = headerIncludeRegex.search(content)
    if match:
        return match.end()
    return 0


def trimExistingMocInclude(content, cppFileName):
    mocStrRegex = re.compile(r'#include "moc_%s\.cpp"\n' % fileNameWithoutExtension(cppFileName))
    match = mocStrRegex.search(content)
    if match:
        return content[:match.start()] + content[match.end():]
    return content


def processFile(root, fileName):
    # pylint: disable=global-statement
    global dirty
    macroFound = hasMacro(root+"/"+fileName)
    logVerbose("Inspecting %s %s" %
               (root+"/"+fileName, "[Has Q_OBJECT / Q_GADGET / Q_NAMESPACE]" if macroFound else ""))

    if macroFound:
        cppFileName = matchingCPPFile(root, fileName)
        logVerbose("  -> %s" % cppFileName)

        if not os.path.exists(cppFileName):
            log("file %s didn't exist (which might not be an error)" % cppFileName)
            return

        if args.replaceExisting or not cppHasMOCInclude(cppFileName):
            dirty = True
            if args.dryRun:
                log("Missing moc include file: %s" % cppFileName)
            else:
                log("Updating %s" % cppFileName)

                with open(cppFileName, "r", encoding="utf8") as f:
                    content = f.read()

                if args.replaceExisting:
                    content = trimExistingMocInclude(content, cppFileName)

                loc = getMocInsertionLocation(cppFileName, content)
                if args.insertAtEnd:
                    with open(cppFileName, "a", encoding="utf8") as f:
                        f.write('\n#include "moc_%s.cpp"\n' % fileNameWithoutExtension(cppFileName))
                else:
                    with open(cppFileName, "w", encoding="utf8") as f:
                        f.write(content[:loc] + ('#include "moc_%s.cpp"\n' %
                                fileNameWithoutExtension(cppFileName)) + content[loc:])


def log(content):
    if not args.quiet:
        print(content)


def logVerbose(content):
    if args.verbose:
        print(content)


################################ MAIN #################################
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="""Script to add inclusion of mocs to files recursively.
        The source files either need to be in the same directories as the header files or in parallel directories,
        where the root of the headers are specified using --header-prefix and the root of the sources are specified using --source-prefix.
        If either header-prefix or source-prefix is the current directory, then they may be omitted.""")
    parser.add_argument("--dry-run", "-n", dest="dryRun", action='store_true', help="only report files to be updated")
    parser.add_argument("--quiet", "-q", dest="quiet", action='store_true', help="suppress output")
    parser.add_argument("--verbose", "-v", dest="verbose", action='store_true')
    parser.add_argument("--header-prefix", metavar="directory", dest="headerPrefix",
                        help="This directory will be replaced with source-prefix when "
                             "searching for matching source files")
    parser.add_argument("--source-prefix", metavar="directory", dest="sourcePrefix", help="see --header-prefix")
    parser.add_argument("--excludes", metavar="directory", dest="excludes", nargs="*",
                        help="directories to be excluded, might either be in the form of a directory name, "
                        "e.g. 3rdparty or a partial directory prefix from the root, e.g 3rdparty/parser")
    parser.add_argument("--insert-at-end", dest="insertAtEnd", action='store_true',
                        help="insert the moc include at the end of the file instead of the beginning")
    parser.add_argument("--replace-existing", dest="replaceExisting", action='store_true',
                        help="delete and readd existing MOC include statements")
    parser.add_argument(dest="root", default=".", metavar="directory",
                        nargs="?", help="root directory for the operation")

    args = parser.parse_args()

    root = args.root
    if args.headerPrefix:
        root += "/" + args.headerPrefix

    path = os.walk(root)
    for root, directories, files in path:
        # Filter out directories specified in --exclude
        directories[:] = [d for d in directories if not shouldExclude(root, d)]

        for file in files:
            if file.endswith(".h") or file.endswith(".hpp"):
                processFile(root, file)

    if not dirty:
        log("No changes needed")

    sys.exit(-1 if dirty else 0)




#!/usr/bin/env python3

import sys
import re
from unidecode import unidecode
from jinja2 import Template


class Emoji(object):
    def __init__(self, code, shortname, unicodename):
        self.code = ''.join(['\\U'+c.rjust(8, '0') for c in code.strip().split(' ')])
        self.shortname = shortname
        self.unicodename = unicodename

def generate_provider_class(**kwargs):
    entrycount = sum([len(c[1]) for c in kwargs.items()])
    tmpl = Template('''\
    // SPDX-FileCopyrightText: Nheko Contributors
    //
    // SPDX-License-Identifier: GPL-3.0-or-later

    // DO NOT EDIT Provider.h DIRECTLY! EDIT IT IN scripts/emoji_codegen.py AND RUN scripts/codegen.sh!

    #pragma once
    #include <array>
    #include "Emoji.h"

    namespace emoji {
    class Provider
    {
    public:
        // all emoji for QML purposes
        static const std::array<Emoji, {{ entrycount }}> emoji;
    };
    } // namespace emoji
    ''')
    d = dict(entrycount=entrycount)
    print(tmpl.render(d))
def generate_qml_list(**kwargs):
    entrycount = sum([len(c[1]) for c in kwargs.items()])
    tmpl = Template('''
    std::array<Emoji, {{ entrycount }} > emoji::Provider::emoji = {
    {%- for c in kwargs.items() %}
    // {{ c[0].capitalize() }}
    {%- for e in c[1] %}
    Emoji{std::u16string_view(u"{{ e.code }}"), std::u16string_view(u"{{ e.shortname }}"), std::u16string_view(u"{{ e.unicodename }}"), emoji::Emoji::Category::{{ c[0].capitalize() }}},
    {%- endfor %}
    {%- endfor %}
};
    ''')
    d = dict(kwargs=kwargs, entrycount=entrycount)
    print(tmpl.render(d))
def usage():
    print('usage: emoji_codegen.py {impl|header} /path/to/emoji-test /path/to/shortcodes.txt')
if __name__ == '__main__':
    if len(sys.argv) < 4:
        usage()
        sys.exit(1)

    mode = sys.argv[1]
    if mode != 'impl' and mode != 'header':
        usage()
        sys.exit(1)
    filename = sys.argv[2]
    shortcodefilename = sys.argv[3]

    people = []
    nature = []
    food = []
    activity = []
    travel = []
    objects = []
    symbols = []
    flags = []

    categories = {
        'Smileys & Emotion': people,
        'People & Body': people,
        'Animals & Nature': nature,
        'Food & Drink': food,
        'Travel & Places': travel,
        'Activities': activity,
        'Objects': objects,
        'Symbols': symbols,
        'Flags': flags,
        'Component': symbols
    }
    shortcodeDict = {}
    # for my sanity - this strips newlines
    for line in open(shortcodefilename, 'r', encoding="utf8"):
        longname, shortname = line.strip().split(':')
        shortcodeDict[longname] = shortname
    current_category = ''
    for line in open(filename, 'r', encoding="utf8"):
        if line.startswith('# group:'):
            current_category = line.split(':', 1)[1].strip()

        if not line or line.startswith('#'):
            continue

        segments = re.split(r'\s+[#;] ', line.strip())
        if len(segments) != 3:
            continue

        code, qualification, charAndName = segments

        # skip unqualified versions of same unicode
        if qualification != 'fully-qualified':
            continue

        char, name = re.match(r'^(\S+) E\d+\.\d+ (.*)$', charAndName).groups()
        shortname = name
        # until skin tone is handled, keep them around
        ## discard skin tone variants for sanity
        # if "skin tone" in name and qualification != 'component': 
        #    continue
        # if qualification == 'component' and not "skin tone" in name:
        #    continue
        #TODO: Handle skintone modifiers in a sane way
        basicallyTheSame = False
        if code in shortcodeDict:
            shortname = shortcodeDict[code]
        else:
            shortname = shortname.lower()
            if shortname.endswith(' (blood type)'):
                shortname = shortname[:-13]
            if shortname.endswith(': red hair'):
                shortname = "red_haired_" + shortname[:-10]
            if shortname.endswith(': curly hair'):
                shortname = "curly_haired_" + shortname[:-12]
            if shortname.endswith(': white hair'):
                shortname = "white_haired_" + shortname[:-12]
            if shortname.endswith(': bald'):
                shortname = "bald_" + shortname[:-6]
            if shortname.endswith(': beard'):
                shortname = "bearded_" + shortname[:-7]
            if shortname.endswith(' face'):
                shortname = shortname[:-5]
            if shortname.endswith(' button'):
                shortname = shortname[:-7]
            if shortname.endswith(' banknote'):
                shortname = shortname[:-9]

            # FIXME: Is there a better way to do this?
            matchobj = re.match(r'^flag: (.*)$', shortname)
            if shortname.startswith("flag: "):
                country = shortname[5:]
                shortname = country + " flag"
            shortname = shortname.replace("u.s.", "us")
            shortname = shortname.replace("&", "and")

            if shortname == name.lower():
                basicallyTheSame = True

            shortname = shortname.replace("-", "_")
            shortname = re.sub(r'\W', '_', shortname)
            shortname, = re.match(r'^_*(.+)_*$', shortname).groups()
            shortname = re.sub(r'_{2,}', '_', shortname)
            shortname = unidecode(shortname)
        # if basicallyTheSame: 
        #    shortname = ""
        categories[current_category].append(Emoji(code, shortname, name))

    # Use xclip to pipe the output to clipboard.
    # e.g ./emoji_codegen.py emoji.json | xclip -sel clip
    # alternatively - delete the var from src/emoji/Provider.cpp, and do ./codegen.sh emojis shortcodes >> ../src/emoji/Provider.cpp
    func = None
    if mode == 'impl':
        func = generate_qml_list
    else:
        func = generate_provider_class
    func(people=people, nature=nature, food=food, activity=activity, travel=travel, objects=objects, symbols=symbols, flags=flags)
